""" Robotica Schedule. """
import asyncio
from typing import Dict, Any, Set, List  # NOQA
import logging

import yaml

from robotica.lifx import Lifx
from robotica.audio import Audio

logger = logging.getLogger(__name__)


Action = Dict[str, Any]


class Executor:
    def __init__(
            self, loop: asyncio.AbstractEventLoop,
            config: str, lifx: Lifx, audio: Audio) -> None:
        self._loop = loop
        with open(config, "r") as file:
            self._config = yaml.safe_load(file)
        self._lifx = lifx
        self._audio = audio

    def is_action_required_for_locations(
            self, locations: Set[str], action: Action) -> bool:
        lights = None
        message = None
        music = None

        if self._lifx.is_action_required_for_locations(locations):

            if 'lights' in action:
                lights = action['lights']

        if self._audio.is_action_required_for_locations(locations):

            if 'message' in action:
                message = action['message']

            if 'music' in action:
                music = action['music']

        return any([lights, message, music])

    async def _do_lights(self, locations: Set[str], action: Action) -> None:
        if 'lights' in action:
            lights = action['lights']
            lifx = self._lifx

            lights_action = lights['action']
            if lights_action == "flash":
                await lifx.flash(locations=locations)
            elif lights_action == "wake_up":
                await lifx.wake_up(locations=locations)
            else:
                logger.error("Unknown action '%s'.", action)

    async def _do_audio(self, locations: Set[str], action: Action) -> None:
        if 'message' in action:
            message = action['message']
            audio = self._audio

            await audio.say(
                locations=locations,
                text=message['text'])

        if 'music' in action:
            music = action['music']
            audio = self._audio

            await audio.music_play(
                locations=locations,
                play_list=music['play_list'])

    async def do_action(self, locations: Set[str], action: Action) -> None:
        if self.is_action_required_for_locations(locations, action):
            await asyncio.gather(
                self._do_lights(locations, action),
                self._do_audio(locations, action),
                loop=self._loop
            )

    async def do_actions(self, locations: Set[str], actions: List[Action]) -> None:
        for action in actions:
            await self.do_action(locations, action)



import asyncio
import base64
import datetime
import logging
from json import JSONDecodeError

import yaml

from aiohttp import web
from typing import Any, Awaitable, Callable, Union

from robotica import __version__ as version
from robotica.executor import Executor
from robotica.schedule import Schedule

logger = logging.getLogger(__name__)

JsonType = Any
Handler = Callable[[int], Awaitable[JsonType]]


class Http:
    def __init__(
            self, loop: asyncio.AbstractEventLoop,
            config: str,
            executor: Executor,
            schedule: Schedule) -> None:
        self._loop = loop
        with open(config, "r") as file:
            self._config = yaml.safe_load(file)
        self._disabled = self._config['disabled']
        self._username = self._config['username']
        self._password = self._config['password']
        self._executor = executor
        self._schedule = schedule

    @staticmethod
    def _get_version(request: web.Request) -> JsonType:
        return {
            'version': version,
        }

    async def _post_execute(self, request: web.Request) -> JsonType:
        try:
            await self._executor.do_actions(request.data['locations'], request.data['actions'])
            return {'status': 'success'}
        except KeyError:
            logger.error("Required value missing.")
            raise web.HTTPBadRequest()

    def _get_schedule(self, request: web.Request) -> JsonType:
        try:
            date = request.match_info['date']
            year, month, day = [int(str) for str in date.split("-")]
            parsed_date = datetime.date(year=year, month=month, day=day)
        except ValueError:
            raise web.HTTPBadRequest()
        schedule = self._schedule.get_schedule_for_date(parsed_date)
        return [s.to_json() for s in schedule]

    def _get_application(self) -> web.Application:
        """ Setup router to point to our handlers. """
        app = web.Application(middlewares=[self._authorize, self._rest])
        app.router.add_get('/version/', self._get_version)
        app.router.add_post('/execute/', self._post_execute)

        schedule = app.router.add_resource('/schedule/{date}/')
        schedule.add_route('GET', self._get_schedule)
        return app

    def start(self) -> None:
        if not self._disabled:
            self._app = self._get_application()
            self._handler = self._app.make_handler()
            f = self._loop.create_server(self._handler, '0.0.0.0', 8080)
            self._srv = self._loop.run_until_complete(f)
            logger.info('serving on %s', self._srv.sockets[0].getsockname())

    def stop(self) -> None:
        if not self._disabled:
            self._srv.close()
            self._loop.run_until_complete(self._srv.wait_closed())
            self._loop.run_until_complete(self._app.shutdown())
            self._loop.run_until_complete(self._handler.shutdown(60.0))
            self._loop.run_until_complete(self._app.cleanup())

    async def _authorize(
            self, app: web.Application, handler: Handler) -> Handler:
        """ Middleware to check the authorization of the request. """
        async def middleware(request: web.Request) -> web.Response:
            """ Middleware handler to check authorization. """
            authorization = request.headers.get('Authorization')
            if authorization is None:
                return web.HTTPForbidden()
            split = authorization.split(' ')
            if len(split) != 2 or split[0] != 'Basic':
                return web.HTTPForbidden()
            credentials = split[1]
            credentials_decoded = base64.b64decode(credentials).decode('ASCII')
            username_password = credentials_decoded.split(':')
            if (len(username_password) != 2
                    or username_password[0] != self._username
                    or username_password[1] != self._password):
                return web.HTTPForbidden()
            return await handler(request)
        return middleware

    async def _rest(self, app: web.Application, handler: Handler) -> Handler:
        """ Middleware will convert data to/from python dictionary and call handler. """
        async def middleware(request: web.Request) -> web.Response:
            """ Middleware handler. """
            if request.method == "GET":
                request.data = request.query_string
            else:
                content_type = request.content_type
                if content_type == "application/json":
                    try:
                        request.data = await request.json()
                    except JSONDecodeError:
                        logger.error("Invalid JSON received")
                        raise web.HTTPBadRequest
                else:
                    logger.error("Unsupported content type '%s'.", request.content_type)
                    return web.HTTPNotAcceptable()

            for accept in request.headers.getall('ACCEPT', []):
                if accept == "application/json":
                    data_out = await handler(request)
                    return web.json_response(data_out)

            logger.error("Unsupported ACCEPT header '%s'.", accept)
            return web.HTTPNotAcceptable()
        return middleware



# -*- coding: utf-8 -*-

"""Console script for Robotica."""
import asyncio
import logging

import click
import click_log

from robotica.executor import Executor
from robotica.lifx import Lifx
from robotica.audio import Audio
from robotica.http import Http
from robotica.schedule import Schedule


logger = logging.getLogger(__name__)


@click.command()
@click.option('--audio', default="config/audio.yaml", help='Path to audio config.')
@click.option('--lifx', default="config/lifx.yaml", help='Path to LIFX config.')
@click.option('--executor', default="config/executor.yaml", help='Path to executor config.')
@click.option('--schedule', default="config/schedule.yaml", help='Path to schedule config.')
@click.option('--http', default="config/http.yaml", help='Path to HTTP config.')
@click_log.simple_verbosity_option()
@click_log.init()
def main(audio: str, lifx: str, executor: str, schedule: str, http: str) -> None:
    """Console script for robotica."""
    loop = asyncio.get_event_loop()

    lifx_obj = Lifx(loop, lifx)
    lifx_obj.start()

    audio_obj = Audio(loop, audio)

    executor_obj = Executor(loop, executor, lifx_obj, audio_obj)

    schedule_obj = Schedule(schedule, executor_obj)
    schedule_obj.start()

    http_obj = Http(loop, http, executor_obj, schedule_obj)
    http_obj.start()

    try:
        loop.run_forever()
    finally:
        http_obj.stop()
        schedule_obj.stop()
        lifx_obj.stop()
        loop.close()



""" Robotica Schedule. """
import datetime
from typing import Dict, List, Set, Any, Optional  # NOQA
import logging

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from dateutil.parser import parse
import yaml
from apscheduler.schedulers.base import BaseScheduler

from robotica.executor import Executor, Action

logger = logging.getLogger(__name__)


_weekdays = {
    'monday': 0,
    'tuesday': 1,
    'wednesday': 2,
    'thursday': 3,
    'friday': 4,
    'saturday': 5,
    'sunday': 6,
}


class TimeEntry:
    def __init__(
            self,
            time: datetime.time,
            locations: Set[str],
            actions: List[Action]) -> None:
        self.time = time
        self.locations = locations
        self.actions = actions

    def to_json(self) -> Dict[str, Any]:
        return {
            'time': str(self.time),
            'locations': list(self.locations),
            'actions': self.actions,
        }

    def __str__(self) -> str:
        return "schedule@%s" % self.time

    def __repr__(self) -> str:
        return "<schedule %s %s %s>" % (
            self.time, self.locations, self.actions)


class Schedule:
    def __init__(self, schedule_path: str, executor: Executor) -> None:
        with open(schedule_path, "r") as file:
            self._schedule = yaml.safe_load(file)
        self._executor = executor

    def start(self) -> None:
        scheduler = AsyncIOScheduler()
        scheduler.start()
        self.add_tasks_to_scheduler(scheduler)

    def stop(self) -> None:
        pass

    def _parse_entry(
            self,
            date: datetime.date,
            locations: Set[str], entry: Dict,
            time_offset: Optional[datetime.time]) -> List[TimeEntry]:
        result = []  # type: List[TimeEntry]

        locations = locations | set(entry.get('locations', []))
        actions = list(entry.get('actions', []))

        time = entry['time']
        hours, minutes = map(int, time.split(':'))
        parsed_time = datetime.time(hour=hours, minute=minutes)

        if time_offset is not None:
            parsed_datetime = datetime.datetime.combine(date, parsed_time)
            delta = datetime.timedelta(
                hours=time_offset.hour, minutes=time_offset.minute)

            required_datetime = parsed_datetime + delta
            if required_datetime.date() != date:
                logger.error(
                    "Skipping time not for date: %s.",
                    required_datetime)

            parsed_time = required_datetime.time()

        if 'template' in entry:
            template_name = entry['template']
            template_result = self._expand_template(
                date=date,
                time=parsed_time,
                locations=locations,
                template_name=template_name,
            )
            result = result + template_result

        actions = [
            action
            for action in actions
            if self._executor.is_action_required_for_locations(
                locations, action)
        ]

        if len(actions) > 0:
            result.append(TimeEntry(
                time=parsed_time,
                locations=locations,
                actions=actions,
            ))

        return result

    def _expand_template(
            self, date: datetime.date, time: datetime.time, locations: Set[str],
            template_name: str) -> List[TimeEntry]:
        result = []  # type: List[TimeEntry]

        template = self._schedule['template'][template_name]
        template_schedule = template['schedule']

        for template_entry in template_schedule:

            entry_result = self._parse_entry(
                date=date,
                locations=locations,
                entry=template_entry,
                time_offset=time,
            )
            result = result + entry_result

        return result

    def get_days_for_date(self, date: datetime.date) -> List[str]:
        results = []  # type: List[str]

        for name, day in self._schedule['day'].items():
            disabled = day.get('disabled', False)
            when = day.get('when')
            match = True

            if disabled:
                match = False
            elif when is not None:
                found_day_of_week = False
                if 'days_of_week' in when:
                    for required_day_of_week in when['days_of_week']:
                        required_value = _weekdays[required_day_of_week.lower()]
                        if date.weekday() == required_value:
                            found_day_of_week = True
                            break
                    if not found_day_of_week:
                        match = False
                if 'dates' in when:
                    found_date = False
                    for date_str in when['dates']:
                        if isinstance(date_str, str) and ' to ' in date_str:
                            split = date_str.split(" to ", maxsplit=1)
                            first_date = parse(split[0]).date()
                            last_date = parse(split[1]).date()
                        elif isinstance(date_str, str):
                            first_date = parse(date_str).date()
                            last_date = first_date
                        else:
                            first_date = date_str
                            last_date = date_str
                        if first_date <= date <= last_date:
                            found_date = True
                    if not found_date:
                        match = False

            if match:
                logger.debug("Adding schedule %s", name)
                results.append(name)

        # We can easily get from schedule -> replaces, but we want
        # to index the reverse relationship.
        replaced_by = {}  # type: Dict[str, List[str]]
        for name in results:
            replaced_by[name] = []
        for name in results:
            replaces_list = self._schedule['day'][name].get('replaces', [])
            for replaces in replaces_list:
                if replaces in replaced_by:
                    replaced_by[replaces].append(name)

        # For every leaf node - that is any node not in danger of being replaced,
        # we can process its replaces.
        n = 0
        while len(replaced_by) > 0 and n < 10:
            n += 1

            for name in list(replaced_by.keys()):

                # Skip entry if already been removed.
                if name not in replaced_by:
                    continue

                # Get the replaced_by list for the entry.
                replaced_by_list = replaced_by[name]

                # Skip node if not leaf node.
                # Node is a leaf node if no schedules are replacing it.
                if len(replaced_by_list) > 0:
                    continue

                # This node is a leaf, therefore it is not getting replaced.
                # As this node is staying, we should process its replaces list.
                replaces_list = self._schedule['day'][name].get('replaces', [])
                for replaces in replaces_list:
                    logger.debug("Replacing schedule %s", replaces)
                    # For every replaces, we should remove all references to this
                    # node.
                    for __, remove_list in replaced_by.items():
                        if replaces in remove_list:
                            remove_list.remove(replaces)
                    # Remove it from the dictionary.
                    if replaces in replaced_by:
                        del replaced_by[replaces]
                    # We also remove it from the results list.
                    if replaces in results:
                        results.remove(replaces)

                # Now we remove the leaf node from dictionary, so we
                # don't process it again.
                del replaced_by[name]

        # if too many loops, probably a loop in the replaces:
        if n >= 10:
            raise RuntimeError("Possible circular loop in replaces")

        return results

    def get_schedule_for_date(self, date: datetime.date) -> List[TimeEntry]:
        result = []  # type: List[TimeEntry]

        days = self.get_days_for_date(date)

        logger.info("Getting schedule for days %s.", days)
        for day in days:
            logger.debug("Adding day '%s' to schedule.", day)
            locations = set(self._schedule['day'][day]['locations'])
            schedule = self._schedule['day'][day]['schedule']

            for entry in schedule:
                entry_result = self._parse_entry(
                    date=date,
                    locations=locations,
                    entry=entry,
                    time_offset=None,
                )
                result = result + entry_result

        result = sorted(result, key=lambda e: e.time)
        return result

    async def do_task(self, entry: TimeEntry) -> None:
        logger.info("%s: Waking up for %s.", datetime.datetime.now(), entry)
        await self._executor.do_actions(entry.locations, entry.actions)

    async def prepare_for_day(self, scheduler: BaseScheduler) -> None:
        logger.info("%s: Updating schedule.", datetime.datetime.now())
        self.add_tasks_to_scheduler(scheduler)

    def add_tasks_to_scheduler(self, scheduler: BaseScheduler) -> None:
        date = datetime.date.today()
        schedule = self.get_schedule_for_date(date)

        scheduler.remove_all_jobs()
        scheduler.add_job(
            self.prepare_for_day, 'cron', hour="00", minute="00",
            kwargs={'scheduler': scheduler}
        )
        for entry in schedule:
            logger.debug("Adding entry '%s' to scheduler.", entry)
            hour = entry.time.hour
            minute = entry.time.minute
            scheduler.add_job(
                self.do_task, 'cron', hour=hour, minute=minute,
                kwargs={'entry': entry}
            )
